
program pascalbeautifier;

{
|                Pascal Program Formatter
|
|   by J. E. Crider, Shell Oil Company, Houston, Texas 77025
|
|   Copyright (c) 1980 by Shell Oil Company.  Permission to
|   copy, modify, and distribute, but not for profit, is
|   hereby granted, provided that this note is included.
|
|   Extended to handle Turbo Pascal version 4.0 - 5.5 and other extensions
|   by David R. Astels, July-October 1989
|   (c) David R. Astels 1989
|
|   This program formats Pascal programs and acceptable program fragments
|   in a way that is configurable using command line options.
|
|   PREPERATION:  The option pair -w+, -w- enable/disable the insertion
|      of a BEGIN END pair around a structured statement that is within
|      another structured statement.  -p- is the default.
|
|   FORMATTING:  Each structured statement that controls a simple
|      statement is placed on a single line, as if it were a simple
|      statement.  Otherwise, each structured statement is formatted
|      in the following pattern (with indentation "indent"):
|
|            XXXXXX header XXXXXXXX
|                XXXXXXXXXXXXXXXXXX
|                XXXXX body XXXXXX
|            XXXXXXXXXXXXXXXXXX
|
|      where the header is one of:
|
|            while <expression> do begin
|            for <control variable> := <for list> do begin
|            with <record variable list> do begin
|            repeat
|            if <expression> then begin
|            else if <expression> then begin
|            else begin
|            case <expression> of
|            <case label list>: begin
|
|      and the last line either begins with UNTIL or ends with END.
|      Other program parts are formatted similarly.  The headers are:
|
|            <program/procedure/function/unit heading>;
|            label
|            const
|            type
|            var
|            begin
|            (various for records and record variants)
|
|   COMMENTS:  Each comment that starts before or on a specified
|      column on an input line (program constant "commthresh") is
|      copied without shifting or reformatting.  Each comment that
|      starts after "commthresh" is reformatted and left-justified
|      following the aligned comment base column ("alcommbase").
|
|   LABELS:  Each statement label is justified to the left margin and
|      is placed on a line by itself.
|
|   SPACES AND BLANK LINES:  Spaces not at line breaks are discarded.
|      Blank lines are copied from the input if they appear
|      between statements (or appropriate declaration units).  A blank
|      line is inserted above each significant part of each program/
|      procedure/function if one is not already there.
|
|   CONTINUATION:  Lines that are too long for an output line are
|      continued with additional indentation ("contindent").
|
|   INPUT FORM:  The program expects as input a program or program
|      fragment in Standard Pascal.  A program fragment is acceptable
|      if it consists of a sequence of (one or more) properly ordered
|      program parts; examples are:  a statement part (that is, a
|      compound statement), or a TYPE part and a VAR part followed by
|      procedure declarations.  If the program fragment is in serious
|      error, then the program may copy the remainder of the input file
|      to the output file without significant modification.  Error
|      messages may be inserted into the output file as comments.
|
|   There are several command line options for PB, see the documentation
|      file (PB.DOC) for information.
|}

uses
    Dos;

const
    hexdigits = ['0'..'9', 'A'..'f', 'a'..'f'];             { hex digits }
    maxrwlen = 32;                                          { size of reserved word strings }
    ordminchar = 32;                                        { ord of lowest char in char set }
    ordmaxchar = 126;                                       { ord of highest char in char set }

    max_builtins = 750;                                     { number of builtin identifiers }

{  Although this program uses the ASCII character set, conversion to
    most other character sets should be straightforward. }

{  The following parameters may be adjusted for the installation: }
    maxinlen = 255;                                         { maximum width of input line + 1 }

type
    natural = 0..maxint;
    inrange = 0..maxinlen;
    outrange = 0..256;
    stringtype = string [maxrwlen];

    errortype = (longline, noendcomm, notquote, longword, notdo, notof, notend, notthen, notbegin, notuntil, notsemicolon,
            notcolon, notparen, noeof, notinterface, notimplementation);

    chartype = (illegal, special, chapostrophe, chleftparen, chleftbracket, chrightparen, chperiod, digit, chcolon, chcomma,
            chsemicolon, chrightbracket, chlessthan, chgreaterthan, letter, chleftbrace, chrightbrace, chequal, chhash,
            chdollar, chcaret);

                                                            { for reserved word recognition }
    resword = (                                             { reserved words ordered by length }
            rwif, rwdo, rwof, rwto, rwin, rwor,             { length: 2 }
            rwend, rwfor, rwvar, rwdiv, rwmod, rwset, rwand, rwnot, rwnil, rwshl, rwshr, rwxor,
                                                            { length: 3 }
            rwthen, rwelse, rwwith, rwgoto, rwcase, rwtype, rwfile, rwunit, rwuses,
                                                            { length: 4 }
            rwbegin, rwuntil, rwwhile, rwarray, rwconst, rwlabel,
                                                            { length: 5 }
            rwrepeat, rwrecord, rwdownto, rwpacked, rwinline, rwstring, rwobject,
                                                            { length: 6 }
            rwprogram, rwforward, rwvirtual,                { length: 7 }
            rwfunction, rwabsolute, rwexternal,             { length: 8 }
            rwprocedure, rwinterface, rwinterrupt,          { length: 9 }
            rwdestructor,                                   { length: 10 }
            rwconstructor,                                  { length: 11 }
            rwimplementation,                               { length: 14 }
            rwx);                                           { length: 15 for table sentinel }
    rwstringtype = string [maxrwlen];

    firstclass = (                                          { class of word if on new line }
            newclause,                                      { start of new clause }
            continue,                                       { continuation of clause }
            alcomm,                                         { start of aligned comment }
            contalcomm,                                     { continuation of aligned comment }
            uncomm,                                         { start of unaligned comment }
            contuncomm,                                     { continuation of unaligned comment }
            stmtlabel,                                      { statement label }
            backindent);                                    { unindent this line a level }
    wordtype = record                                       { data record for word }
        whenfirst: firstclass;                              { class of word if on new line }
        puncfollows: boolean;                               { to reduce dangling punctuation }
        blanklncount: natural;                              { number of preceding blank lines }
        spaces: integer;                                    { number of spaces preceding word }
        base: - 9..maxinlen;                                { inlin.buf[base] precedes word }
        size: inrange
    end;                                                    { length of word in inlin.buf }

    symboltype = (                                          { symbols for syntax analysis }
            semicolon, sybegin, syend,                      { three insertable symbols first }
            syif, sydo, syof, sythen, syelse, sygoto, sycase, syuntil, syrepeat, syrecord, forwhilewith, progprocfunc,
            declarator, syvar, otherword, othersym, leftparen, rightparen, period, leftbracket, rightbracket, sysubrange,
            intconst, colon, comma, ident, comment, equal, syeof, commentend, procdirect, hash, dollar, caret, syunit,
            syinterface, syimplementation, syinline, syobject, syvirtual);
    inserttype = semicolon..syend;
    symbolset = set of symboltype;
{ *** NOTE: set size of 0..26 REQUIRED for symbolset! }

var
{ command line switch settable variables }

    tab_width: integer;                                     { tab with for input }
    use_builtins: boolean;                                  { special treatment of builtins }
    rw_case: boolean;                                       { reserved word case, false = lower, true = upper }
    initmargin: integer;                                    { initial output margin }
    maxoutlen: integer;                                     { maximum output line length }
    commthresh: integer;                                    { column threshhold in input for comments to be aligned }
    alcommbase: integer;                                    { aligned comments in output start AFTER this column }
    indent: integer;                                        { RECOMMENDED indentation increment }
    contindent: integer;                                    { continuation indentation, >indent }
    endspaces: integer;                                     { number of spaces to precede 'END' }
    commindent: integer;                                    { comment continuation indentation }
    nl_on_end: boolean;                                     { start new line for END }
    proc_indent: boolean;                                   { indent after proc heading }
    sep_paramlist: boolean;                                 { seperate func/proc name and parameter list }
    indent_of: boolean;                                     { nl and indent followinf OF }
    wrap_struct: boolean;                                   { insert begin-end pair around struct stmts in a struct stmt }
    snuggle_ifs: boolean;                                   { snuggle stmts after THEN & ELSE up to the THEN or ELSE }
    snuggle_dos: boolean;                                   { snuggle stmts after DO up to the DO }
    infile: text;                                           { input file }
    outfile: text;                                          { output file }

    in_place: boolean;                                      { true if formatting begin done 'in place' }
    param_list: boolean;                                    { true if parsing in a parameter list }
    last_was_space: boolean;                                { last character output was a space }
    is_rw: boolean;                                         { current symbol is a reserved Turbo Pascal word }
    is_builtin: boolean;                                    { current symbol is a builtin Turbo Pascal identifier }
    num_builtins: integer;                                  { number of builtin identifiers }
    endcomment: boolean;                                    { signal whether parsing comment or not }
    in_directive: boolean;                                  { parsing a compiler directive }
    inlin: record                                           { input line data }
        endoffile: boolean;                                 { end of file on input? }
        ch: char;                                           { current char, buf[index] }
        index: inrange;                                     { subscript of current char }
        len: natural;                                       { length of input line in buf }
                                                            { string ';BEGINEND' in buf[-8..0] }
        buf: array [- 8..256] of char
    end;
    outline: record                                         { output line data }
        blanklns: natural;                                  { number of preceding blank lines }
        len: outrange;                                      { number of chars in buf }
        buf: array [1..256] of char
    end;
    cur_word: wordtype;                                     { current word }
    margin: outrange;                                       { left margin }
    lnpending: boolean;                                     { new line before next symbol? }
    symbol: symboltype;                                     { current symbol }
    prev_sym: symboltype;                                   { previous symbol }

{ Structured Constants }
    headersyms: symbolset;                                  { headers for program parts }
    strucsyms: symbolset;                                   { symbols that begin structured
                                                               statements }
    stmtbeginsyms: symbolset;                               { symbols that begin statements }
    stmtendsyms: symbolset;                                 { symbols that follow statements }
    stopsyms: symbolset;                                    { symbols that stop expression scan }
    recendsyms: symbolset;                                  { symbols that stop record scan }
    datawords: symbolset;                                   { to reduce dangling punctuation }
    noleadspace: symbolset;                                 { symbols that shouldn't have a leading space }
    newword: array [inserttype] of wordtype;
    instring: packed array [1..9] of char;
    firstrw: array [1..maxrwlen] of resword;
    rwword: array [rwif..rwimplementation] of rwstringtype;
    rwsy: array [rwif..rwimplementation] of symboltype;

    builtins: array [1..max_builtins] of rwstringtype;

    charclass: array [char] of chartype;
{ above is portable form; possible ASCII form is: }
{    charclass: array [' '..'~'] of chartype;     }
    symbolclass: array [chartype] of symboltype;





{|
 |  Compare two strings, without regard to case
 |
 |  Entry:  strings to compare
 |  Return: 0 if they are equal
 |          -1 if s1 < s2
 |          1 if s1 > s2
 |}

function StriCmp(var s1, s2: string): integer;

var
    caps1, caps2: string;
    i: integer;

begin
    caps1 := s1;
    for i := 1 to Length(caps1) do
        caps1[i] := UpCase(caps1[i]);

    caps2 := s2;
    for i := 1 to Length(caps2) do
        caps2[i] := UpCase(caps2[i]);

    if caps1 < caps2 then
        StriCmp := - 1
    else if caps1 = caps2 then
        StriCmp := 0
    else
        StriCmp := 1;
end;



procedure MakeBuiltins(name: string);

var
    start, i, j: integer;
    datafile: text;
    out_form: string;
    path: string;
    temp: rwstringtype;

begin
    if not use_builtins then                                { don't load if builtins not wanted }
        Exit;

    path := FSearch(name, GetEnv('PATH'));
    if path = '' then begin
        Writeln('Can''t find builtin file: ', name);
        Halt(1);
    end;

    Assign(datafile, path);
    Reset(datafile);

    start := num_builtins;

    while not Eof(datafile) and (num_builtins < max_builtins) do begin
        Readln(datafile, out_form);
        Inc(num_builtins);
        builtins[num_builtins] := out_form;
    end;

    if num_builtins >= max_builtins then begin
        Writeln('Builtin array too small!');
        Halt;
    end;
    Close(datafile);

{ sort the array (insertion sort) if this wasn't the default file }

    if name <> 'BUILTINS.PB' then
        for i := start + 1 to num_builtins do begin
            j := i;
            temp := builtins[i];

            while StriCmp(builtins[j - 1], temp) > 0 do begin
                builtins[j] := builtins[j - 1];
                Dec(j);
            end;

            builtins[j] := temp;
        end;

end;                                                        { make-builtins }



procedure strucconsts;                                      { establish values of structured
                                                               constants }

var
    i: ordminchar..ordmaxchar;                              { loop index }
    ch: char;                                               { loop index }

procedure buildinsert(symbol: inserttype; inclass: firstclass; inpuncfollows: boolean; inspaces, inbase: integer; insize:
        inrange);

begin
    with newword[symbol] do begin
        whenfirst := inclass;
        puncfollows := inpuncfollows;
        blanklncount := 0;
        spaces := inspaces;
        base := inbase;
        size := insize
    end;
end;                                                        { buildinsert }

procedure buildrw(rw: resword; symword: string; symbol: symboltype);

var
    i: integer;

begin
    rwword[rw][0] := Chr(Length(symword));
    for i := 1 to Length(symword) do
        rwword[rw][i] := symword[i];                        { reserved word string }
    rwsy[rw] := symbol;                                     { map to symbol }
end;                                                        { buildrw }




begin                                                       { strucconsts }
                                                            { symbol sets for syntax analysis }
    headersyms := [progprocfunc, declarator, syvar, sybegin, syeof, syinterface, syimplementation];
    strucsyms := [sycase, syrepeat, syif, forwhilewith];
    stmtbeginsyms := strucsyms + [sybegin, ident, sygoto, syinline];
    stmtendsyms := [semicolon, syend, syuntil, syelse, syeof];
    stopsyms := headersyms + strucsyms + stmtendsyms + [sygoto];
    recendsyms := [rightparen, syend, syeof, progprocfunc];

    datawords := [otherword, intconst, ident, syend];
    noleadspace := [semicolon, rightparen, period, comment, caret, rightbracket, sysubrange, colon, comma];

                                                            { words for insertable symbols }
    buildinsert(semicolon, continue, False, 0, - 9, 1);
    buildinsert(sybegin, continue, False, 1, - 8, 5);
    if nl_on_end then
        buildinsert(syend, backindent, True, endspaces, - 3, 3)
    else
        buildinsert(syend, newclause, True, endspaces, - 3, 3);

    instring := ';beginend';

                                                            { constants for recognizing reserved
                                                               words }
    firstrw[1] := rwif;                                     { length: 1 }
    firstrw[2] := rwif;                                     { length: 2 }
    buildrw(rwif, 'IF', syif);
    buildrw(rwdo, 'DO', sydo);
    buildrw(rwof, 'OF', syof);
    buildrw(rwto, 'TO', othersym);
    buildrw(rwin, 'IN', othersym);
    buildrw(rwor, 'OR', othersym);
    firstrw[3] := rwend;                                    { length: 3 }
    buildrw(rwend, 'END', syend);
    buildrw(rwfor, 'FOR', forwhilewith);
    buildrw(rwvar, 'VAR', syvar);
    buildrw(rwdiv, 'DIV', othersym);
    buildrw(rwmod, 'MOD', othersym);
    buildrw(rwset, 'SET', othersym);
    buildrw(rwand, 'AND', othersym);
    buildrw(rwnot, 'NOT', othersym);
    buildrw(rwshl, 'SHL', othersym);
    buildrw(rwshr, 'SHR', othersym);
    buildrw(rwxor, 'XOR', othersym);
    buildrw(rwnil, 'NIL', otherword);
    firstrw[4] := rwthen;                                   { length: 4 }
    buildrw(rwthen, 'THEN', sythen);
    buildrw(rwelse, 'ELSE', syelse);
    buildrw(rwwith, 'WITH', forwhilewith);
    buildrw(rwgoto, 'GOTO', sygoto);
    buildrw(rwcase, 'CASE', sycase);
    buildrw(rwtype, 'TYPE', declarator);
    buildrw(rwunit, 'UNIT', syunit);
    buildrw(rwuses, 'USES', declarator);
    buildrw(rwfile, 'FILE', othersym);
    firstrw[5] := rwbegin;                                  { length: 5 }
    buildrw(rwbegin, 'BEGIN', sybegin);
    buildrw(rwuntil, 'UNTIL', syuntil);
    buildrw(rwwhile, 'WHILE', forwhilewith);
    buildrw(rwarray, 'ARRAY', othersym);
    buildrw(rwconst, 'CONST', declarator);
    buildrw(rwlabel, 'LABEL', declarator);
    firstrw[6] := rwrepeat;                                 { length: 6 }
    buildrw(rwrepeat, 'REPEAT', syrepeat);
    buildrw(rwrecord, 'RECORD', syrecord);
    buildrw(rwdownto, 'DOWNTO', othersym);
    buildrw(rwpacked, 'PACKED', othersym);
    buildrw(rwinline, 'INLINE', syinline);
    buildrw(rwstring, 'STRING', othersym);
    buildrw(rwobject, 'OBJECT', syobject);
    firstrw[7] := rwprogram;                                { length: 7 }
    buildrw(rwprogram, 'PROGRAM', progprocfunc);
    buildrw(rwforward, 'FORWARD', procdirect);
    buildrw(rwvirtual, 'VIRTUAL', syvirtual);
    firstrw[8] := rwfunction;                               { length: 8 }
    buildrw(rwfunction, 'FUNCTION', progprocfunc);
    buildrw(rwabsolute, 'ABSOLUTE', othersym);
    buildrw(rwexternal, 'EXTERNAL', procdirect);
    firstrw[9] := rwprocedure;                              { length: 9 }
    buildrw(rwprocedure, 'PROCEDURE', progprocfunc);
    buildrw(rwinterface, 'INTERFACE', syinterface);
    buildrw(rwinterrupt, 'INTERRUPT', procdirect);
    firstrw[10] := rwdestructor;
    buildrw(rwdestructor, 'DESTRUCTOR', progprocfunc);
    firstrw[11] := rwconstructor;
    buildrw(rwconstructor, 'CONSTRUCTOR', progprocfunc);
    firstrw[12] := rwimplementation;
    firstrw[13] := rwimplementation;
    firstrw[14] := rwimplementation;                        { length: 14 }
    buildrw(rwimplementation, 'IMPLEMENTATION', syimplementation);
    firstrw[15] := rwx;                                     { length: 15 for table sentinel }


                                                            { constants for lexical scan }
    for i := ordminchar to ordmaxchar do begin
        charclass[Chr(i)] := illegal
    end;
    for ch := 'a' to 'z' do begin                           { !!! implementation-dependent!  (but
                                                               can be replaced with 52 explicit
                                                               assignments) }
        charclass[ch] := letter;
        charclass[UpCase(ch)] := letter
    end;
    for ch := '0' to '9' do
        charclass[ch] := digit;
    charclass['#'] := chhash;                               { to make #9.. work }
    charclass['_'] := letter;
    charclass[' '] := special;
    charclass['$'] := chdollar;                             { to make $FFFF.. work }
    charclass[''''] := chapostrophe;
    charclass['('] := chleftparen;
    charclass[')'] := chrightparen;
    charclass['['] := chleftbracket;
    charclass[']'] := chrightbracket;
    charclass['*'] := special;
    charclass['+'] := special;
    charclass['-'] := special;
    charclass['.'] := chperiod;
    charclass['/'] := special;
    charclass[':'] := chcolon;
    charclass[','] := chcomma;
    charclass[';'] := chsemicolon;
    charclass['<'] := chlessthan;
    charclass['='] := chequal;
    charclass['>'] := chgreaterthan;
    charclass['@'] := special;
    charclass['^'] := chcaret;
    charclass['{'] := chleftbrace;
    charclass['}'] := chrightbrace;
    symbolclass[illegal] := othersym;
    symbolclass[special] := othersym;
    symbolclass[chcaret] := caret;
    symbolclass[chdollar] := dollar;
    symbolclass[chhash] := hash;
    symbolclass[chapostrophe] := otherword;
    symbolclass[chleftparen] := leftparen;
    symbolclass[chrightparen] := rightparen;
    symbolclass[chleftbracket] := leftbracket;
    symbolclass[chrightbracket] := rightbracket;
    symbolclass[chperiod] := period;
    symbolclass[digit] := intconst;
    symbolclass[chcolon] := colon;
    symbolclass[chcomma] := comma;
    symbolclass[chsemicolon] := semicolon;
    symbolclass[chlessthan] := othersym;
    symbolclass[chgreaterthan] := othersym;
    symbolclass[letter] := ident;
    symbolclass[chleftbrace] := comment;
    symbolclass[chrightbrace] := commentend;
    symbolclass[chequal] := equal;

end;                                                        { strucconsts }

{ writeline/writeerror/readline convert between files and lines. }

procedure writeline;                                        { write buffer into output file }

var
    i: outrange;                                            { loop index }

begin
    with outline do begin
        while blanklns > 0 do begin
            Writeln(outfile);
            blanklns := blanklns - 1
        end;
        if len > 0 then begin
            for i := 1 to len do
                Write(outfile, buf[i]);
            Writeln(outfile);
            len := 0
        end
    end;
end;                                                        { writeline }

procedure writeerror(error: errortype);                     { report error to output }

var
    i, ix: inrange;                                         { loop index, limit }

begin
    writeline;
    Write(outfile, ' (*  !!! error, ');
    case error of
        longline: Write(outfile, 'shorter line');
        noendcomm: Write(outfile, 'end of comment');
        notquote: Write(outfile, 'final "''" on line');
        longword: Write(outfile, 'shorter word');
        notdo: Write(outfile, '"do"');
        notof: Write(outfile, '"of"');
        notend: Write(outfile, '"end"');
        notthen: Write(outfile, '"then"');
        notbegin: Write(outfile, '"begin"');
        notuntil: Write(outfile, '"until"');
        notsemicolon: Write(outfile, '";"');
        notcolon: Write(outfile, '":"');
        notparen: Write(outfile, '")"');
        noeof: Write(outfile, 'end of file');
        notinterface: Write(outfile, 'interface');
        notimplementation: Write(outfile, 'implementation');
    end;
    Write(outfile, ' expected');
    if error >= longword then begin
        Write(outfile, ', not "');
        with inlin, cur_word do begin
            if size > maxrwlen then
                ix := maxrwlen
            else
                ix := size;
            for i := 1 to ix do
                Write(outfile, buf[base + i])
        end;
        Write(outfile, '"')
    end;
    if error = noeof then
        Write(outfile, ', FORMATTING STOPS');
    Writeln(outfile, ' !!!  *)');
end;                                                        { writeerror }

procedure readline;                                         { read line into input buffer }

var
    c: char;                                                { input character }
    nonblank: boolean;                                      { is char other than space? }

begin
    with inlin do begin
        len := 0;
        if Eof(infile) then
            endoffile := True
        else begin                                          { get next line }
            while not Eoln(infile) do begin
                Read(infile, c);
                if (c < ' ') then begin
                    if (c = #9) and (not endcomment) then   { handle tabs }
                        while (len + 1) mod tab_width > 0 do begin
                            len := len + 1;
                            if len < maxinlen then
                                buf[len] := ' ';
                        end;
                    if not ((c = #12) and (len > 0)) then
                        c := ' ';                           { replace control chars with space }
                end;                                        { except leading FF }
                len := len + 1;
                if len < maxinlen then
                    buf[len] := c
            end;
            Readln(infile);
            if len >= maxinlen then begin                   { input line too long }
                writeerror(longline);
                len := maxinlen - 1
            end;
            nonblank := False;
            repeat                                          { trim line }
                if len = 0 then
                    nonblank := True
                else if buf[len] <> ' ' then
                    nonblank := True
                else
                    len := len - 1
            until nonblank
        end;
        len := len + 1;                                     { add exactly ONE trailing blank }
        buf[len] := ' ';
        index := 0
    end;
end;                                                        { readline }

{ startword/finishword/copyword convert between lines and words.
   auxiliary procedures getchar/nextchar precede. }

procedure getchar;                                          { get next char from input buffer }

begin
    with inlin do begin
        index := index + 1;
        ch := buf[index]
    end;
end;                                                        { getchar }

function nextchar: char;                                    { look at next char in input buffer }

begin
    with inlin do
        nextchar := buf[index + 1];
end;                                                        { nextchar }

procedure startword(startclass: firstclass);                { note beginning of word, and count
                                                               preceding lines and spaces }

var
    first: boolean;                                         { is word the first on input line? }

begin
    first := False;
    with inlin, cur_word do begin
        whenfirst := startclass;
        blanklncount := 0;
        while (index >= len) and not endoffile do begin
            if len = 1 then
                blanklncount := blanklncount + 1;
            if startclass = contuncomm then
                writeline
            else
                first := endcomment;
            readline;                                       { with exactly ONE trailing blank }
            getchar;
            if ch = Chr(12) then begin
                writeline;
                Writeln(outfile, Chr(12));
                blanklncount := 0;
                getchar
            end;
        end;
        spaces := 0;                                        { count leading spaces }
        if not endoffile then begin
            while ch = ' ' do begin
                getchar;
                if not endcomment then
                    spaces := spaces + 1;
            end
        end;
        if first then
            spaces := 1;
        base := index - 1
    end;
end;                                                        { startword }

procedure finishword;                                       { note end of word }

begin
    with inlin, cur_word do begin
        puncfollows := (symbol in datawords) and (ch <> ' ');
        size := index - base - 1
    end;
end;                                                        { finishword }

procedure copyword(newline: boolean; cur_word: wordtype);   { copy word from input buffer into
                                                               output buffer }

var
    i: integer;                                             { outline.len excess, loop index }

begin
    with cur_word, outline do begin
        i := maxoutlen - len - spaces - size;
        if newline or (i < 0) or ((i = 0) and puncfollows) or ((symbol = syend) and nl_on_end) then
            writeline;
        if len = 0 then begin                               { first word on output line }
            blanklns := blanklncount;
            case whenfirst of                               { update LOCAL word.spaces }
                newclause: spaces := margin;
                continue: spaces := margin + contindent;
                alcomm: spaces := alcommbase;
                contalcomm: spaces := alcommbase + commindent;
                uncomm:;                                    { spaces := spaces }
                contuncomm:;                                { spaces := spaces }
                stmtlabel: spaces := initmargin;
                backindent: spaces := margin - indent
            end;
            if spaces + size > maxoutlen then begin
                spaces := maxoutlen - size;                 { reduce spaces }
                if spaces < 0 then begin
                    writeerror(longword);
                    size := maxoutlen;
                    spaces := 0
                end
            end
        end else if endcomment then begin
            if (prev_sym = equal) or (sep_paramlist and (symbol = leftparen) and (prev_sym in [syinline, ident])) or not (
                    symbol = syend) and not (symbol in [comment, commentend]) and not last_was_space and not (symbol in
                    noleadspace) and not (prev_sym in [period, caret]) and not ((symbol in [leftparen, leftbracket]) and (
                    prev_sym in [syinline, ident, rightbracket])) and not ((symbol = intconst) and (prev_sym = hash)) and not
                    (prev_sym in [leftparen, leftbracket, sysubrange]) then
                spaces := 1
        end;

        for i := 1 to spaces do begin                       { put out spaces }
            len := len + 1;
            buf[len] := ' '
        end;

        for i := 1 to size do begin                         { copy actual word }
            len := len + 1;
            buf[len] := inlin.buf[base + i]
        end;
        last_was_space := buf[len] = ' ';
    end;

    prev_sym := symbol;
end;                                                        { copyword }

{ docomment/copysymbol/insert/getsymbol/findsymbol convert between
   words and symbols. }

procedure docomment;                                        { copy aligned or unaligned comment }

procedure copycomment(commclass: firstclass; commbase: inrange);
                                                            { copy words of comment }

begin
    endcomment := False;
    with cur_word do begin                                  { copy comment begin symbol }
        whenfirst := commclass;
        spaces := commbase - outline.len;
        copyword((spaces < 0) or (blanklncount > 0), cur_word)
    end;
    commclass := Succ(commclass);
    symbol := othersym;
    if inlin.ch = '$' then begin                            { check for directive }
        startword(commclass);
        getchar;
        finishword;
        in_directive := True;
        copyword(False, cur_word);
    end;
    with inlin do begin
        repeat                                              { loop for successive words }
            startword(commclass);
            endcomment := endoffile;                        { premature end? }
            if endcomment then
                writeerror(noendcomm)
            else begin
                repeat
                    if ch = '*' then begin
                        getchar;
                        if ch = ')' then begin
                            Dec(index);
                            finishword;
                            copyword(False, cur_word);
                            startword(commclass);
                            getchar;
                            endcomment := True;
                            in_directive := False;
                            symbol := commentend;
                            getchar
                        end
                    end else if ch = '}' then begin
                        finishword;
                        copyword(False, cur_word);
                        startword(commclass);
                        endcomment := True;
                        in_directive := False;
                        symbol := commentend;
                        getchar
                    end else
                        getchar
                until (ch = ' ') or endcomment
            end;
            finishword;
            copyword(False, cur_word);
        until endcomment
    end;
end;                                                        { copycomment }

begin                                                       { docomment }
    if cur_word.base < commthresh then begin                { copy comment without alignment }
        copycomment(uncomm, cur_word.base)
    end else begin                                          { align and format comment }
        copycomment(alcomm, alcommbase)
    end;
end;                                                        { docomment }

procedure copysymbol(symbol: symboltype; cur_word: wordtype);
                                                            { copy word(s) of symbol }

begin
    if symbol = comment then begin
        docomment;                                          { NOTE: docomment uses global word! }
        lnpending := True
    end else if symbol = semicolon then begin
        copyword(False, cur_word);
        lnpending := not param_list;
    end else begin
        copyword(lnpending, cur_word);
        lnpending := False
    end;
end;                                                        { copysymbol }

procedure Insert(newsymbol: inserttype);                    { copy word for inserted symbol into
                                                               output buffer }

var
    old_sym: symboltype;

begin
    old_sym := symbol;
    symbol := newsymbol;
    copysymbol(newsymbol, newword[newsymbol]);
    symbol := old_sym;
end;                                                        { insert }

procedure getsymbol;                                        { get next non-comment symbol }

procedure findsymbol;                                       { find next symbol in input buffer }

var
    chclass: chartype;                                      { classification of leading char }

procedure checkresword;                                     { check if current identifier is
                                                               reserved word/symbol }

var
    rw, rwbeyond: resword;                                  { loop index, limit }
    all_caps, symword: rwstringtype;                        { copy of symbol word }
    i, b, u, l: integer;                                    { loop index }

begin
    with cur_word, inlin do begin
        size := index - base - 1;
        if size < maxrwlen then begin
            symword[0] := Chr(size);
            for i := 1 to size do
                symword[i] := UpCase(buf[base + i]);
            rw := firstrw[size];
            rwbeyond := firstrw[size + 1];
            symbol := semicolon;

            repeat                                          { check for a reserved word }
                if rw >= rwbeyond then
                    symbol := ident
                else if symword = rwword[rw] then
                    symbol := rwsy[rw]
                else
                    rw := Succ(rw)
            until symbol <> semicolon;
            is_rw := symbol <> ident;
            if is_rw then
                for i := 1 to size do begin
                    if not rw_case then
                        symword[i] := Chr(Ord(symword[i]) + 32);
                    buf[base + i] := symword[i];            { make word lowercase }
                end;

            if use_builtins and not is_rw then begin        { check for builtin identifier }
                l := 1;
                u := num_builtins;
                is_builtin := False;
                while (l <= u) and not is_builtin do begin
                    b := (l + u) div 2;

                    all_caps := builtins[b];
                    for i := 1 to Length(all_caps) do
                        all_caps[i] := UpCase(all_caps[i]);

                    if symword < all_caps then
                        u := b - 1
                    else if symword > all_caps then
                        l := b + 1
                    else
                        is_builtin := True;
                end;
                if is_builtin then
                    for i := 1 to size do
                        buf[base + i] := builtins[b][i];
            end;

            if symbol in [syend, syuntil] then begin
                if spaces < endspaces then
                    spaces := endspaces;
                if nl_on_end then
                    whenfirst := backindent
                else
                    whenfirst := newclause;
            end
        end
    end;
end;                                                        { checkresword }

procedure getname;

begin
    while (charclass[inlin.ch] in [letter, digit]) or (charclass[inlin.ch] = chperiod) and (charclass[nextchar] in [letter,
            digit]) do
        getchar;
    checkresword;
end;                                                        { getname }

procedure getnumber;

begin
    with inlin do begin
        while charclass[ch] = digit do
            getchar;
        if ch = '.' then begin                              { thanks to A.H.J.Sale, watch for
                                                               '..' }
            if charclass[nextchar] = digit then begin       { NOTE: nextchar is a function! }
                symbol := otherword;
                getchar;
                while charclass[ch] = digit do
                    getchar
            end
        end;
        if UpCase(ch) = 'E' then begin
            symbol := otherword;
            getchar;
            if (ch = '+') or (ch = '-') then
                getchar;
            while charclass[ch] = digit do
                getchar
        end
    end;
end;                                                        { getnumber }



procedure gethex;                                           { get hexadecimal number }

begin
    with inlin do
        while (ch in hexdigits) do
            getchar;
end;                                                        { getnumber }

procedure getstringliteral;

var
    endstring: boolean;

begin
    with inlin do begin
        endstring := False;
        repeat
            if ch = '''' then begin
                getchar;
                if ch = '''' then
                    getchar
                else
                    endstring := True
            end else if index >= len then begin             { error, final "'" not on line }
                writeerror(notquote);
                symbol := syeof;
                endstring := True
            end else
                getchar
        until endstring
    end;
end;                                                        { getstringliteral }

begin                                                       { findsymbol }
    startword(continue);
    with inlin do begin
        if endoffile then
            symbol := syeof
        else begin
            chclass := charclass[ch];
            symbol := symbolclass[chclass];
            getchar;                                        { second char }

            case chclass of
                chsemicolon, chrightparen, chleftbrace, special, illegal:;
                letter: getname;
                digit: getnumber;
                chdollar: gethex;
                chapostrophe: getstringliteral;
                chcomma: symbol := comma;
                chcolon: begin
                    if ch = '=' then begin
                        symbol := othersym;
                        getchar
                    end
                end;
                chlessthan: begin
                    if (ch = '=') or (ch = '>') then
                        getchar
                end;
                chgreaterthan: begin
                    if ch = '=' then
                        getchar
                end;
                chleftparen: begin
                    if ch = '*' then begin
                        symbol := comment;
                        getchar
                    end
                end;
                chleftbracket: symbol := leftbracket;
                chrightbracket: symbol := rightbracket;
                chequal: symbol := equal;
                chperiod: begin
                    if ch = '.' then begin
                        symbol := sysubrange;
                        getchar
                    end
                end
            end
        end
    end;
    finishword;
end;                                                        { findsymbol }

begin                                                       { getsymbol }
    repeat
        copysymbol(symbol, cur_word);                       { copy word for symbol to output }
        findsymbol;                                         { get next symbol }
        if lnpending and (symbol = syvirtual) then
            lnpending := False;
    until symbol <> comment;
end;                                                        { getsymbol }

{ block performs recursive-descent syntax analysis with symbols,
   adjusting margin, lnpending, cur_word.whenfirst, and
   cur_word.blanklncount.  auxiliary procedures precede. }

procedure startclause;                                      { (this may be a simple clause, or
                                                               the start of a header) }

begin
    if (symbol in [syend, syuntil]) and nl_on_end then
        cur_word.whenfirst := backindent
    else
        cur_word.whenfirst := newclause;
    lnpending := True;
end;                                                        { startclause }

procedure passsemicolons;                                   { pass consecutive semicolons }

begin
    while symbol = semicolon do begin
        getsymbol;
        startclause
    end;                                                    { new line after ';' }
end;                                                        { passsemicolons }

procedure startpart(need_blank: boolean);                   { start program part }

begin
    with cur_word do begin
        if (blanklncount = 0) and need_blank then
            blanklncount := 1
    end;
    startclause;
end;                                                        { startpart }

procedure startbody(needindent: boolean);                   { finish header, start body of
                                                               structure }

begin
    passsemicolons;
    if needindent then                                      { allow not indenting var... }
        margin := margin + indent;
    startclause;
end;                                                        { startbody }

procedure finishbody(needindent: boolean);

begin
    if needindent then
        margin := margin - indent;
end;                                                        { finishbody }

procedure passphrase(finalsymbols: symbolset);              { process symbols until one of significant
                                                               symbols encountered }

var
    endsyms: symbolset;                                     { complete set of stopping symbols }

begin
    if symbol <> syeof then begin
        endsyms := stopsyms + finalsymbols;
        repeat
            getsymbol
        until symbol in endsyms
    end;
end;                                                        { passphrase }

procedure expect(expectedsym: symboltype; error: errortype; syms: symbolset);

begin
    if symbol = expectedsym then
        getsymbol
    else begin
        writeerror(error);
        while not (symbol in [expectedsym] + syms) do
            getsymbol;
        if symbol = expectedsym then
            getsymbol
    end;
end;                                                        { expect }

procedure dolabel;                                          { process statement label }

var
    nextfirst: firstclass;                                  { (pass whenfirst to statement) }

begin
    with cur_word do begin
        nextfirst := whenfirst;
        whenfirst := stmtlabel;
        lnpending := True;
        getsymbol;
        expect(colon, notcolon, stopsyms);
        whenfirst := nextfirst;
        lnpending := True
    end;
end;                                                        { dolabel }

procedure block(needbody: boolean);                         { process block }

{ needbody is true if procedures/functions need anything beyond a heading
      as in the interface part of a unit }

procedure Heading;                                          { process heading for program,
                                                               procedure, or function }

procedure matchparens;                                      { process parentheses in heading }

begin
    getsymbol;
    while not (symbol in recendsyms) do begin
        if symbol = leftparen then
            matchparens
        else
            getsymbol
    end;
    expect(rightparen, notparen, stopsyms + recendsyms);
end;                                                        { matchparens }

begin                                                       { heading }
    getsymbol;
    passphrase([leftparen]);
    if symbol = leftparen then begin
        param_list := True;
        matchparens;
        param_list := False;
    end;
    if symbol = colon then
        passphrase([semicolon]);
    expect(semicolon, notsemicolon, stopsyms);
end;                                                        { heading }

procedure statement;                                        { process statement }

forward;

procedure stmtlist;                                         { process sequence of statements }

begin
    repeat
        statement;
        passsemicolons
    until symbol in stmtendsyms;
end;                                                        { stmtlist }

procedure compoundstmt(                                     { process compound statement }
        stmtpart: boolean);                                 { statement part of block? }

begin
    getsymbol;
    startbody(True);                                        { new line, indent after 'BEGIN' }
    stmtlist;
    expect(syend, notend, stmtendsyms);
    finishbody(True);                                       { left-indent after 'END' }
end;                                                        { compoundstmt }

procedure statement;                                        { process statement }

procedure checkcompound;                                    { if structured then force compound }

begin
    if symbol = intconst then
        dolabel;
    if (symbol in strucsyms) and wrap_struct then begin     { force compound }
        Insert(sybegin);
        startbody(True);                                    { new line, indent after 'BEGIN' }
        statement;
        Insert(syend);
        finishbody(True);
    end else
        statement;
end;                                                        { checkcompound }

procedure ifstmt;                                           { process if statement }

var
    new_body: boolean;

begin
    passphrase([sythen]);
    expect(sythen, notthen, stopsyms);

    new_body := not (snuggle_ifs or ((symbol in strucsyms) and wrap_struct) or (symbol = sybegin));
    if new_body then
        startbody(True);

    if symbol <> sybegin then
        checkcompound
    else
        statement;

    if new_body then
        finishbody(True);

    if symbol = syelse then begin
        if (prev_sym = syend) and not nl_on_end then begin
            finishbody(True);
            startclause;
        end else if (prev_sym <> syend) then
            startclause;                                    { new line before 'ELSE' unless following 'END' }
        getsymbol;

        if symbol = syif then
            ifstmt
        else begin
            new_body := not (snuggle_ifs or ((symbol in strucsyms) and wrap_struct) or (symbol = sybegin));
            if new_body then
                startbody(True);

            if symbol <> sybegin then
                checkcompound
            else
                statement;

            if new_body then
                finishbody(True);
        end;
    end;
end;                                                        { ifstmt }

procedure repeatstmt;                                       { process repeat statement }

begin
    getsymbol;
    startbody(True);                                        { new line, indent after 'REPEAT' }
    stmtlist;
    startclause;                                            { new line before 'UNTIL' }
    expect(syuntil, notuntil, stmtendsyms);
    finishbody(True);
    passphrase([semicolon]);
end;                                                        { repeatstmt }

procedure fwwstmt;                                          { process for, while, or with
                                                               statement }

var
    new_body: boolean;

begin
    passphrase([sydo]);
    expect(sydo, notdo, stopsyms);

    new_body := not (snuggle_dos or ((symbol in strucsyms) and wrap_struct) or (symbol = sybegin));
    if new_body then
        startbody(True);

    if symbol <> sybegin then
        checkcompound
    else
        statement;

    if new_body then
        finishbody(True);

end;                                                        { fwwstmt }

procedure casestmt;                                         { process case statement }

begin
    passphrase([syof]);
    expect(syof, notof, stopsyms);
    startbody(True);                                        { new line, indent after 'OF' }
    repeat
        if symbol = syelse then
            getsymbol
        else begin
            passphrase([colon]);
            expect(colon, notcolon, stopsyms);
        end;
        if indent_of then
            startbody(True);

        if symbol <> sybegin then
            checkcompound
        else
            statement;

        if indent_of then
            finishbody(True);
        passsemicolons
    until symbol in stopsyms - [syelse];
    expect(syend, notend, stmtendsyms);
    finishbody(True);                                       { left-indent after 'END' }
end;                                                        { casestmt }

begin                                                       { statement }
    if symbol = intconst then
        dolabel;
    if symbol in stmtbeginsyms then begin
        case symbol of
            sybegin: compoundstmt(False);
            sycase: casestmt;
            syif: ifstmt;
            syrepeat: repeatstmt;
            forwhilewith: fwwstmt;
            syinline, ident, sygoto: passphrase([semicolon])
        end
    end;
    if not (symbol in stmtendsyms) then begin
        writeerror(notsemicolon);                           { ';' expected }
        passphrase([semicolon])
    end;
end;                                                        { statement }

procedure passfields(forvariant: boolean);

forward;

procedure dorecord;                                         { process record declaration }

begin
    getsymbol;
    startbody(True);
    passfields(False);
    expect(syend, notend, recendsyms);
    finishbody(True);
end;                                                        { dorecord }

procedure doobject;

begin
    getsymbol;
    if symbol = leftparen then begin                        { handle heritage if present }
        passphrase([rightparen]);
        getsymbol;
    end;

    startbody(True);
    passfields(False);                                      { process instance variable declarations }
    block(False);                                           { process method declarations }
    expect(syend, notend, recendsyms);
    finishbody(True);
end;

procedure dovariant;                                        { process (case) variant part }

begin
    passphrase([syof]);
    expect(syof, notof, stopsyms);
    startbody(True);
    passfields(True);
    finishbody(True);
end;                                                        { dovariant }

procedure doparens(forvariant: boolean);                    { process parentheses in record }

begin
    getsymbol;
    if forvariant then
        startbody(True);
    passfields(False);
    lnpending := False;                                     { for empty field list }
    expect(rightparen, notparen, recendsyms);
    if forvariant then
        finishbody(True);
end;                                                        { doparens }

procedure passfields;                                       { process declarations }
{     procedure passfields (forvariant: boolean); }

begin                                                       { passfields }
    while not (symbol in recendsyms) do begin
        if symbol = semicolon then
            passsemicolons
        else if symbol = syrecord then
            dorecord
        else if symbol = sycase then
            dovariant
        else if symbol = leftparen then
            doparens(forvariant)
        else
            getsymbol
    end;
end;                                                        { passfields }


procedure do_decl;

begin
    startpart(True);                                        { label, const, type, var }
    getsymbol;
    startbody(True);
    repeat
        passphrase([syobject, syrecord]);
        if symbol = syrecord then
            dorecord;
        if symbol = syobject then
            doobject;
        if symbol = semicolon then
            passsemicolons
    until symbol in headersyms;
    finishbody(True)
end;


begin                                                       { block }
    while symbol in [declarator, syvar] do
        do_decl;

    while symbol in [progprocfunc, syvar] do begin
        if symbol = syvar then                              { allow variable declarations between functions }
            do_decl

        else begin                                          { program, procedure, function }
            startpart(needbody);
            Heading;
            if symbol = syvirtual then begin                { handle virtual methods }
                passphrase([semicolon]);
                passsemicolons;
            end;
            if needbody then begin
                startbody(proc_indent);
                if symbol in headersyms then
                    block(True)
                else if symbol in [procdirect, syinline] then begin
                    startpart(False);                       { directive: forward, etc. }
                    passphrase([semicolon]);
                    passsemicolons
                end else
                    writeerror(notbegin);
                finishbody(proc_indent);
            end;
        end;
    end;

    if symbol = syunit then begin
        startpart(True);
        Heading;
        startbody(False);
        if symbol <> syinterface then
            writeerror(notinterface)
        else begin
            startpart(True);
            getsymbol;
        end;
        block(False);                                       { get interface part }
        if symbol <> syimplementation then
            writeerror(notimplementation)
        else begin
            startpart(True);
            getsymbol;
        end;
        block(True);
        if symbol <> syeof then                             { if the block didn't have a statement part }
            passphrase([period]);                           {  grab the trailing END }
        getsymbol;
    end;

    if symbol = sybegin then begin
        startpart(True);                                    { statement part }
        compoundstmt(True);
        if symbol in [sysubrange, period] then
            symbol := semicolon;                            { treat final period as semicolon }
        passsemicolons
    end;
end;                                                        { block }

procedure copyrem;                                          { copy remainder of input }

begin
    writeerror(noeof);
    with inlin do begin
        repeat
            copyword(False, cur_word);
            startword(contuncomm);
            if not endoffile then begin
                repeat
                    getchar
                until ch = ' '
            end;
            finishword;
        until endoffile
    end;
end;                                                        { copyrem }

procedure initialize;                                       { initialize global variables }

var
    i: 1..9;                                                { loop index }

begin
    with inlin do begin
        for i := 1 to 9 do
            buf[i - 9] := instring[i];                      { string ';BEGINEND' in buf[-8..0] }
        endoffile := False;
        ch := ' ';
        index := 0;
        len := 0
    end;
    with outline do begin
        blanklns := 0;
        len := 0
    end;
    with cur_word do begin
        whenfirst := contuncomm;
        puncfollows := False;
        blanklncount := 0;
        spaces := 0;
        base := 0;
        size := 0
    end;
    margin := initmargin;
    lnpending := False;
    symbol := othersym;
    prev_sym := othersym;
    endcomment := True;
    in_directive := False;
    param_list := False;
end;                                                        { initialize }



{|
 |  Process a command line option
 |
 |  Entry:  body of option (less the - or /)
 |}

procedure ProcessOption(Str: string);

var
    temp, vcode: integer;

begin
    case UpCase(Str[1]) of
        'A': begin
            Val(Copy(Str, 2, 7), temp, vcode);
            if vcode = 0 then
                alcommbase := temp;
        end;
        'B': use_builtins := Str[2] = '+';
        'C': case UpCase(Str[2]) of
            'C': begin
                Val(Copy(Str, 3, 7), temp, vcode);
                if vcode = 0 then
                    commindent := temp;
            end;
            'I': begin
                Val(Copy(Str, 3, 7), temp, vcode);
                if vcode = 0 then
                    contindent := temp;
            end;
            'T': begin
                Val(Copy(Str, 3, 7), temp, vcode);
                if vcode = 0 then
                    commthresh := temp;
            end;
        end;
        'E': begin
            Val(Copy(Str, 2, 7), temp, vcode);
            if vcode = 0 then
                endspaces := temp;
        end;
        'F': sep_paramlist := Str[2] = '+';
        'I': begin
            Val(Copy(Str, 2, 7), temp, vcode);
            if vcode = 0 then
                indent := temp;
        end;
        'L': begin
            Val(Copy(Str, 2, 7), temp, vcode);
            if vcode = 0 then
                maxoutlen := temp;
        end;
        'M': begin
            Val(Copy(Str, 2, 7), temp, vcode);
            if vcode = 0 then
                initmargin := temp;
        end;
        'N': nl_on_end := Str[2] = '+';
        'O': indent_of := Str[2] = '+';
        'P': proc_indent := Str[2] = '+';
        'R': rw_case := Str[2] = '+';
        'S': case UpCase(Str[2]) of
            'D': snuggle_dos := Str[3] = '+';
            'I': snuggle_ifs := Str[3] = '+';
            '+': begin
                snuggle_dos := True;
                snuggle_ifs := True;
            end;
            '-': begin
                snuggle_dos := False;
                snuggle_ifs := False;
            end;
        end;

        'T': begin
            Val(Copy(Str, 2, 7), temp, vcode);
            if vcode = 0 then
                tab_width := temp;
        end;
        'W': wrap_struct := Str[2] = '+';
    end
end;



{|
 |  Process a response file
 |
 |  Entry:  name of response file
 |}

procedure ProcessResponseFile(name: string);

var
    infile: text;
    Str: string;

begin
    if Pos('.', name) = 0 then
        name := name + '.PB';

    Assign(infile, name);
    Reset(infile);

    while not Eof(infile) do begin
        Readln(infile, Str);

        if Str[1] = '@' then
            ProcessResponseFile(Copy(Str, 2, 79))

        else if Str[1] = '#' then
            MakeBuiltins(Copy(Str, 2, 79))

        else
            ProcessOption(Str);
    end;

    Close(infile);
end;



{ set switch defaults and process command line options }

procedure set_switches;

var
    dummy: word;
    i, start_param: integer;
    Str: string;
    bakname, name, dir, ext: string;
    def_path: PathStr;

begin
    num_builtins := 0;
    tab_width := 4;
    use_builtins := True;
    rw_case := False;
    initmargin := 0;
    maxoutlen := 125;
    commthresh := 4;
    alcommbase := 60;
    indent := 4;
    contindent := 8;
    endspaces := 3;
    commindent := 3;
    nl_on_end := True;
    proc_indent := False;
    sep_paramlist := False;
    indent_of := False;
    in_place := False;
    wrap_struct := False;
    snuggle_dos := False;
    snuggle_ifs := False;

    Assign(infile, '');
    Reset(infile);
    Assign(outfile, '');
    Rewrite(outfile);

    MakeBuiltins('BUILTINS.PB');                            { load default builtins }

    def_path := FSearch('DEFAULT.PB', GetEnv('PATH'));      { load default options }
    if def_path <> '' then
        ProcessResponseFile(def_path);

    if ParamCount > 0 then begin
        start_param := 1;

        Str := ParamStr(1);
        if not (Str[1] in ['-', '/', '@', '#']) then begin  { check for input file }
            Close(infile);
            Close(outfile);
            Assign(infile, FExpand(Str));
            Reset(infile);
            Inc(start_param);
            FSplit(FExpand(Str), dir, name, ext);
            bakname := dir + name + '.BAK';

            Str := ParamStr(2);
            if (ParamCount > 1) and not (Str[1] in ['-', '/', '@', '#']) then begin
                                                            { check for output file }
                Assign(outfile, FExpand(Str));
                Inc(start_param);
                in_place := False;
            end else begin                                  {$I-}
                Assign(outfile, bakname);                   { erase old .BAK file }
                Erase(outfile);                             {$I+}
                dummy := IOResult;
                Close(infile);
                Rename(infile, bakname);                    { rename input file to .BAK }
                Assign(infile, bakname);
                Reset(infile);
                Assign(outfile, FExpand(ParamStr(1)));      { open new file of original name }
                in_place := True;
            end;

            Rewrite(outfile);
        end;

        for i := start_param to ParamCount do begin
            Str := ParamStr(i);

            if Str[1] in ['-', '/'] then                    { check for option }
                ProcessOption(Copy(Str, 2, 79))

            else if Str[1] = '@' then                       { check for response file }
                ProcessResponseFile(Copy(Str, 2, 79))

            else if Str[1] = '#' then                       { check for user builtin file }
                MakeBuiltins(Copy(Str, 2, 79));
        end
    end
end;



begin                                                       { pascalbeautifier }
    set_switches;
    strucconsts;
    initialize;
{  ***************  Files may be opened here. }
    getsymbol;
    block(True);
    if not inlin.endoffile then
        copyrem;
    writeline;
    Close(infile);
    Close(outfile);
end.                                                        { pascalbeautifier }
